'***************************************************************************************************
'*  Name    : PRTOS24_Demo.BAS                                                                     *
'*  Author  : John Barrat                                                                          *
'*  Notice  : Copyright (c) 2017 John Barrat                                                       *
'*          : All Rights Reserved                                                                  *
'*  Date    : 09/08/2017                                                                           *
'*  Version : 1.0                                                                                  *
'*  Notes   : This Demo program is intended to demonstrate all of the main functions of PRTOS24    *
'*          : For the purposes of the demo the target circuit uses a 4 channel Scope and an serial * 
'*          : Terminal.  This has been tested in simulation on ISIS. CCT diag available on request *
'*          : On-start up the demo will offer up the menu below.                                   *
'*                                                                                                 *
'*          P-RTOS24 Test Program...                                                               *
'*                                                                                                 *
'*          1    Start/Stop Yielding Task                                                          *
'*          2    Start/Stop Delayed Task                                                           *
'*          3    Start/Stop Binary Semaphore Task                                                  *
'*          4    Start/Stop Cyclic Timer                                                           *
'*                                                                                                 *
'*          5    Event Timeout Test                                                                *
'*          6    Serial Buffer Test                                                                *
'*          7    Inter task Message Test                                                           *
'*                                                                                                 *
'*          8    Show Tick Count                                                                   *
'*          L    List Active Tasks and TaskIDs                                                     *
'*                                                                                                 *
'*          9    Test Invalid TaskID Error                                                         *
'*          A    Test Task Blocking Error                                                          *
'*          V    Report Version Number                                                             *
'*                                                                                                 *
'*          Hit Spacebar to return to menu at any time                                             *
'*                                                                                                 *
'*          Command>                                                                               *
'*                                                                                                 *
'*          Traces will appear on the scope representing various tasks as they are started or      *
'*          Stopped from the first 4 menu items. The remaining menu items will output to the       *
'*          Virtual terminal.                                                                      *
'*                                                                                                 *
'***************************************************************************************************
;-------------------------------------------------------------------------------
;**** Added by Fuse Configurator ****
; Use the Fuse Configurator plug-in to change these settings

    Device = 24FJ64GA002

    Config Config2 = FCKSM_CSDCMD, FNOSC_PRI, I2C1SEL_PRI, IESO_ON, IOL1WAY_ON, OSCIOFNC_OFF, POSCMOD_HS, SOSCSEL_SOSC, WUTSEL_LEG
    Config Config1 = BKBUG_OFF, COE_OFF, FWDTEN_OFF, FWPSA_PR128, GCP_OFF, GWRP_OFF, ICS_PGx1, JTAGEN_ON, WDTPOST_PS32768, WINDIS_ON

;**** End of Fuse Configurator Settings ****
;-------------------------------------------------------------------------------

    Declare Xtal = 64
    Declare Hserial_Baud = 19200
    Declare HRSOut1_Pin = PORTB.14
    Declare HRSIn_Pin = PORTB.15
    Declare Optimiser_Level = 3
    Declare Create_Coff On

    PPS_Input(cIn_Pin_RP15, cIn_Fn_U1RX)    ; UART1  Receive (U1RX)  Connected to RP15
    PPS_Output(cOut_Pin_RP14, cOut_Fn_U1TX) ; UART1 Transmit Connected to RP14
    PPS_Output(cOut_Pin_RP8, cOut_Fn_NULL)  ; default port pin Connected to RP8

'P-RTOS24 Configuration Settings -------------------------------------------------------------------
     
    $define OSTASKS_COUNT 14               ' Maximum Number of Tasks  
    $define OSPRIO_COUNT 3                 ' Number of priority levels

' Configure Interrupts
    $define OSENABLE_INTERRUPTS            ' Enable interrupts - required for Timer services
      $define OSENABLE_TIMER True          ' Enables timer service
      $define OSENABLE_TIMEOUTS True       ' allow timeouts for events and counters
      $define OSTICK_SOURCE T1             ' T1, T2, EXT
      $define OSTIMER_PRESCALE 8           ' 256, 56, 8 or 1 (Off)
      $define OSTIMER_PRELOAD $9C40        ' Preload value 100Hz Clock (10mSec period) (64Mhz Clock)
'      $define OSTIMER_PRELOAD $4E20        ' Preload value 100Hz Clock (10mSec period) (32Mhz Clock)      
      $define OSTICK_CTR_SIZE 4            ' Size of OS Tick Counter (bytes) (must be 0, 2 or 4 max)
      $define OSENABLE_CYCLIC_TIMERS True  ' allow cyclic timers to be created

' Configure Events
    $define OSENABLE_EVENTS                ' Enables Events
      $define OSEVENTS_COUNT 5             ' Max number of events
      $define OSENABLE_SEMAPHORES True     ' Event Semaphores enabled
      $define OSENABLE_MESSAGES True       ' Event Messages enabled

' Configure Serial IO                                                                                               
    $define OSENABLE_SERIO_INTERRUPT       ' Introduces OS_WaitRXData
      $define OSENABLE_BUFFERED_SERIO True ' buffer sizes must be a power of 2. 
        $define cUSART1_RxBufferSize 32    ' Set up the buffer size for input 
        $define cUSART1_TxBufferSize 64    ' and output circular buffers

' Configure Error Handler
    $define OSENABLE_ERROR_HANDLER         ' Enable exception handling
      $define OSWATCHDOG_TIMEOUT 20        ' Tick count before OSWatchdog triggers
      $define OSUSER_ERROR_HANDLER         ' Override default handler

Include "PRTOS24Main.inc"                  ' Bring in the P24RTOS 

    Dim t_Count As Word
    Dim t_BinSem As Word
    Dim e_BinSem As Word
    Dim c_Timer As Word
    Dim t_Task9 As Word
    Dim t_Task10 As Word
    Dim t_Task11 As Word
    Dim t_Delay As Word
    Dim t_TOTest As Word
    Dim t_CharOut As Word
    Dim t_MenuTask As Word
    Dim t_RecTask As Word
    Dim t_MsgTask As Word
    Dim e_MsgTask As Word
    Dim e_MsgReply As Word
    Dim t_ShowTicks As Word
    Dim t_BadTask As Word

    TRISB = $C000   ' Port B to output except for 14 and 15 which are connected to USART1
    
    GoTo Main       ' Skip over tasks
    
    Task9:          ' Yielding Task  Blue Trace
    Do
    High PORTB.9
    DelayUS 50
    Low PORTB.9
    OS_Yield()
    Loop

    Task10:         ' Yielding Task  Magenta trace
    Do
    High PORTB.10
    DelayUS 50
    Low PORTB.10
    OS_Yield()
    Loop
    
    Task11:         '  Run Once Task Green Trace
    Dim Ctr11 As WREG12
    Do
    Ctr11 = 0
    Low PORTB.11
    Do
    Toggle PORTB.11
    Inc Ctr11
    OS_Yield()
    Loop Until Ctr11 = $3F
    Low PORTB.11
    OS_Stop()
    Loop

    CountTsk:       ' This task runs continuously counting and signalling every $FF cycles
    Dim Ctr As WREG12.LowByte
    Do
    Inc Ctr
    If Ctr = $FF Then
      OSSignalSemaphore(e_BinSem)
    EndIf
    OS_Yield()
    Loop

    DelayedTest:    ' This task Starts task9 for 1 OS Tick and stops task9 for 2 OS Ticks
    Do
    OSStartTask(t_Task9, OSNO_DELAY)
    OS_Delay(1)
    OSStopTask(t_Task9)
    OS_Delay(2)
    Loop

    BinSemTask:     ' This task waits on a semaphore from the Count task then starts a Task11 as a one shot
    Low PORTB.11
    Do
    OS_WaitSem(e_BinSem, OSNO_TIMEOUT)
    OSStartTask(t_Task11, OSNO_DELAY)
    OS_Delay(2)
    Loop

    CyclicTimer1:
    Toggle PORTB.11
    OS_CycReturn()

    Dim CharIn As Byte
    CharOutTask:
    Do
    HRSOut  "\rSerial Buffer Test...\r\r",_
    "Key presses will be echoed to the screen\r"
    Ch:
    OS_WaitRXData(OSNO_TIMEOUT)
    HRSIn CharIn
    If CharIn = " " Then
      OS_ReplaceTask(MenuOutTask, 2)
    Else
      HRSOut CharIn
    EndIf
    GoTo Ch                            ' try again if chars remaining will return immediately
    Loop

    CharOutTOTask:
    Dim i As Byte
    Do
    HRSOut  "\rTimeout test...\r\r"
    ChRep:
    HRSOut  "Press a key within 10 Secs\r"
    ChLp:
    OS_WaitRXData(250)                 ' Set 2.5 Sec timeout
    If OSTimeout() = True Then
      Inc i
      Select i
      Case 1
        HRSOut Rep $08\8,"7.5 Secs"
        GoTo ChLp
      Case 2
        HRSOut Rep $08\8,"5 Secs  "
        GoTo ChLp
      Case 3
        HRSOut Rep $08\8,"2.5 Secs"
        GoTo ChLp
      Case 4
        HRSOut "\rTimed Out - restarting test\r\r"
        i = 0
        GoTo ChRep
      EndSelect
    Else
      i = 0
      HRSIn CharIn
      If CharIn = " " Then
        OS_ReplaceTask(MenuOutTask, 2)
      Else
        HRSOut Rep $08\8, "10 Secs "
      EndIf
    EndIf
    GoTo ChLp
    Loop

    ShowTickCount:
    HRSOut "\rTick Count $", Hex4 OSGetTicks()
    Do
    OS_WaitRXData(10)     ' wait 10 ticks before updating
    If OSTimeout() = True Then
'      OS_WaitTXEmpty()
      HRSOut Rep $08\5, "$", Hex4 OSGetTicks()  ' backspace to start of line and print tick count
    Else
      HRSIn CharIn
      If CharIn = " " Then OS_ReplaceTask(MenuOutTask, 2)
    EndIf
    Loop

    SendMessageTask:
    Dim MsgStr[8] As Byte
    Dim MsgAddr As Dword
    Do
    HRSOut "\rMessage Event Test...\r\r"
'    OS_WaitTXEmpty()
    HRSOut "Enter up to 8 chars followed by Enter\r"
    e_MsgTask = OSCreateMessage(OSNO_MESSAGE)
    e_MsgReply = OSCreateBinSem(False)
    t_RecTask = OSCreateTask(ReceiveMessageTask, 3, True)
    i = 0
    Do
    MsgAddr = OSReadMessage(e_MsgTask)
    OS_WaitRXData(OSNO_TIMEOUT)
    HRSIn CharIn
    If i = 0 And CharIn = " " Then
      OSDestroyTask(t_RecTask)
      OSDestroyEvent(e_MsgTask)
      If e_MsgTask <> OSNO_EVENT Then HRSOut "failed to delete e_MsgTask\r"
      OSDestroyEvent(e_MsgReply)
      If e_MsgReply <> OSNO_EVENT Then HRSOut "failed to delete e_MsgReply\r"
      OS_ReplaceTask(MenuOutTask, 2)
    Else
      If CharIn = $0D Then SigMsg
      MsgStr[i] = CharIn
      Inc i
      If i >= SizeOf MsgStr Then
        SigMsg:
        HRSOut "Sent - ", Str MsgStr\8
        OSSignalMessage(e_MsgTask, MsgStr)
        HRSOut " - Waiting Msg Received Event\r"
        OS_WaitSem(e_MsgReply, OSNO_TIMEOUT)     ' Now wait until ReceiveMessage Task has completed
        Clear MsgStr
        Clear i
        HRSOut "\rSend another message or hit spacebar to return to main menu\r"
      EndIf
    EndIf
    Loop
    Loop

    ReceiveMessageTask:
    Dim j As Byte
    Do
    OS_WaitMessage(e_MsgTask, OSNO_TIMEOUT)       ' returns with WREG2.dword holding address of message
    HRSOut "Received - "
    OSReadMessage(e_MsgTask)
    j = 0
    Do
    CharIn = Ptr8(WREG12 ++)
    If CharIn = 0 Then Break
    HRSOut CharIn
    Inc j
    Loop Until j = 8
    HRSOut " - Signalling Msg Received Event"
    OSSignalSemaphore(e_MsgReply)                ' tell SendMessage task you have done all printing
    Loop
    
    BadTask:                                     ' This  task will never return to the OS
    Do
      DelayMS 2
    Loop

    MenuOutTask:
    Dim TmpDW As Dword
    Do
    CharIn = 0
    HRSOut  "\rP-RTOS24 Test Program...\r\r"
    OS_WaitTXEmpty()
    HRSOut  "1    Start/Stop Yielding Task\r"
    OS_WaitTXEmpty()
    HRSOut  "2    Start/Stop Delayed Task\r"
    OS_WaitTXEmpty()
    HRSOut  "3    Start/Stop Binary Semaphore Task\r"
    OS_WaitTXEmpty()
    HRSOut  "4    Start/Stop Cyclic Timer\r\r"
    OS_WaitTXEmpty()
    HRSOut  "5    Event Timeout Test\r"
    OS_WaitTXEmpty()
    HRSOut  "6    Serial Buffer Test\r"
    OS_WaitTXEmpty()
    HRSOut  "7    Inter task Message Test\r\r"
    OS_WaitTXEmpty()
    HRSOut  "8    Show Tick Count\r"
    OS_WaitTXEmpty()
    HRSOut  "L    List Active Task and Event ID's\r\r"
    OS_WaitTXEmpty()
    HRSOut  "9    Test Invalid TaskID Error\r"
    OS_WaitTXEmpty()
    HRSOut  "A    Test Task Blocking Error\r"
    OS_WaitTXEmpty()
    HRSOut  "V    Report version Number\r"
    OS_WaitTXEmpty()
    HRSOut  "\rHit Spacebar to return to menu at any time\r\r"
    OS_WaitTXEmpty()
    HRSOut  "Command>  "
    ch1:
    OS_WaitRXData(OSNO_TIMEOUT)               ' Wait for Input Interrupt
    HRSIn CharIn
    HRSOut $08
    HRSOut CharIn
    Select CharIn
    Case "1"
      If OSTaskRunning(t_Task10) = True Then
        OSStopTask(t_Task10)
      Else
        OSStartTask(t_Task10, OSNO_DELAY)
      EndIf
    Case "2"
      If OSTaskRunning(t_Task9) = True Then OSStopTask(t_Task9)
      If OSTaskRunning(t_Delay) = True Then
        OSStopTask(t_Delay)
      Else
        OSStartTask(t_Delay, OSNO_DELAY)
      EndIf
    Case "3"
      If t_BinSem = OSNO_TASK Then
        e_BinSem = OSCreateBinSem(False)
        t_Count = OSCreateTask(CountTsk, 2, True)
        t_BinSem = OSCreateTask(BinSemTask, 2, True)
      Else
        OSStopTask(t_BinSem)
        OSStopTask(t_Count)
        OSDestroyTask(t_BinSem)
        OSDestroyEvent(e_BinSem)
        OSDestroyTask(t_Count)
      EndIf
    Case "4"
      If c_Timer = OSNO_TASK Then
        c_Timer = OSCreateCycTmr(CyclicTimer1, 0, 5, OSCT_CONTINUOUS)
      Else
        OSStopTask(c_Timer)
        OSDestroyTask(c_Timer)
      EndIf
    Case "5"
      OS_ReplaceTask(CharOutTOTask, 3)
    Case "6"
      OS_ReplaceTask(CharOutTask, 3)
    Case "7"
      OS_ReplaceTask(SendMessageTask, 3)
    Case "8"
      OS_ReplaceTask(ShowTickCount, 1)
    Case "9"
      OSStartTask(OSTASKS_COUNT + 1, OSNO_DELAY)     ' Start a task using an invalid ID
    Case "A"
      If t_BadTask = OSNO_TASK Then
        t_BadTask = OSCreateTask(BadTask, 1, True)
      Else
        OSStartTask(t_BadTask, OSNO_DELAY)
      EndIf
    Case "V"
      TmpDW = OSVerNo()
      HRSOut "\rMajor Ver ", Hex2 TmpDW.Byte3,_
             " Minor Ver ", Hex2 TmpDW.Byte2,_
             " Release ", Hex2 TmpDW.Byte1,_
             " Build ", Hex2 TmpDW.Byte0, "\rCommand>  "
    Case "L"
      HRSOut "\rTask Name   TaskID\r"
      If t_Count <> OSNO_TASK Then  HRSOut "t_Count     $" , Hex2 t_Count , "\r"
      If t_BinSem <> OSNO_TASK Then HRSOut "t_BinSem    $" , Hex2 t_BinSem , "\r"
      If c_Timer <> OSNO_TASK Then HRSOut "c_Timer     $" , Hex2 c_Timer , "\r"
      If t_Task9 <> OSNO_TASK Then HRSOut "t_Task9     $" , Hex2 t_Task9 , "\r"
      If t_Task10 <> OSNO_TASK Then HRSOut "t_Task10    $" , Hex2 t_Task10 , "\r"
      If t_Task11 <> OSNO_TASK Then HRSOut "t_Task11    $" , Hex2 t_Task11 , "\r"
      If t_Delay <> OSNO_TASK Then HRSOut "t_Delay     $" , Hex2 t_Delay , "\r"
      If t_TOTest <> OSNO_TASK Then HRSOut "t_TOTest    $" , Hex2 t_TOTest , "\r"
      If t_CharOut <> OSNO_TASK Then HRSOut "t_CharOut   $" , Hex2 t_CharOut  , "\r"
      If t_MenuTask <> OSNO_TASK Then HRSOut "t_MenuTask  $" , Hex2 t_MenuTask  , "\r"
      If t_RecTask <> OSNO_TASK Then HRSOut "t_RecTask   $" , Hex2 t_RecTask  , "\r"
      If t_MsgTask <> OSNO_TASK Then HRSOut "t_MsgTask   $" , Hex2 t_MsgTask  , "\r"
      If t_ShowTicks <> OSNO_TASK Then HRSOut "t_ShowTicks $" , Hex2 t_ShowTicks  , "\r"
      If t_BadTask <> OSNO_TASK Then HRSOut "t_BadTask   $" , Hex2 t_BadTask , "\r"
      
      HRSOut "\rEvent Name  EventID\r"
      If e_MsgTask <> OSNO_TASK Then HRSOut "e_MsgTask   $" , Hex2 e_MsgTask  , "\r"
      If e_MsgReply <> OSNO_TASK Then HRSOut "e_MsgReply  $" , Hex2 e_MsgReply  , "\r"
      If e_BinSem <> OSNO_TASK Then HRSOut "e_BinSem    $" , Hex2 e_BinSem  , "\r"
      HRSOut  "Command>  "
        
    Case " "
      GoTo MenuOutTask
    EndSelect
    GoTo ch1
    Loop
    
    Proc UserErrorProc()
      If OSState = True Then OS_WaitTXEmpty()
      HRSOut "\rOriginator $", Hex2 LastError.HighByte, " Error $", Hex2 LastError.LowByte, "\rCommand>  "         
    EndProc

    Main:
    OSInit()
    HRSOut "                    \r"       '  Seem to need this to clear out crud in Virtual Terminal

    '  Create a few tasks
    t_Task9 = OSCreateTask(Task9, 2, False)
    t_Task10 = OSCreateTask(Task10, 2, False)
    t_Task11 = OSCreateTask(Task11, 2, False)
    t_Delay = OSCreateTask(DelayedTest, 2, False)
    t_CharOut = OSCreateTask(CharOutTask, 2, False)
    t_TOTest = OSCreateTask(CharOutTOTask, 2, False)
    t_MenuTask = OSCreateTask(MenuOutTask, 2, True)
    
    ' These tasks/events will be created and destroyed as required in the running program
    c_Timer = OSNO_TASK
    t_MsgTask = OSNO_TASK
    t_BinSem = OSNO_TASK
    t_Count = OSNO_TASK
    t_ShowTicks = OSNO_TASK
    t_BadTask = OSNO_TASK
    e_BinSem = OSNO_EVENT
                               
    OSRun()                              ' run the OS - this will not return

    Stop
