'****************************************************************
'*  Name    : PRTOS24Main.inc                                   *
'*  Author  : John Barrat                                       *
'*  Notice  : Copyright (c) 2017 John Barrat                    *
'*          : All Rights Reserved                               *
'*  Date    : 09/08/2017                                        *
'*  Version : 1.0                                               *
'*  Notes   :                                                   *
'*          :                                                   *
'****************************************************************
' Version Info -------------------------------------------------------------------------------------
    $define OSMajorVer 1
    $define OSMinorVer 1
    $define OSRelease  1
    $define oSBuild    2

' Some basic definitions ---------------------------------------------------------------------------

    $ifndef True
      $define True 1                       ' Use for setting Boolean value
    $endif
    $ifndef False
      $define False 0                      ' Use for clearing a boolean Value
    $endif
    $define IsTrue > 0
     
' Minimum configuration   
    $ifndef OSTASKS_COUNT
      $define OSTASKS_COUNT 6              ' Maximum Number of Tasks
    $endif                       
    
    $ifndef OSPRIO_COUNT
      $define OSPRIO_COUNT 3               ' Number of priority levels
    $endif

' Configure Interrupts
   $ifdef OSENABLE_INTERRUPTS
      $ifndef OSENABLE_TIMER 
        $define OSENABLE_TIMER True        ' Enables timer service
        $define OSENABLE_TIMEOUTS True     ' allow timeouts for events and counters
        $define OSTICK_SOURCE T1           ' T1, T2, EXT
        $define OSTIMER_PRESCALE 8         ' 256, 56, 8 or 1 (Off)
'        $define OSTIMER_PRELOAD $4E20      ' Preload value 100Hz Clock (10mSec period) (32Mhz Clock)
        $define OSTIMER_PRELOAD $9C40      ' Preload value 100Hz Clock (10mSec period) (64Mhz Clock)
      $endif
   $endif

' Configure Events
    $ifdef OSENABLE_EVENTS
      $ifndef OSEVENTS_COUNT
        $define OSEVENTS_COUNT 5             ' Max number of events 
      $endif
      $define OSENABLE_SEMAPHORES True     ' Event Semaphores enabled
      $define OSENABLE_MESSAGES True       ' Event Messages enabled
      '$define OSENABLE_EVENT_FLAGS False  ' Event Flags enabled - not yet implemented
      '$define OSEVENT_FLAGS 1             ' Max Event flags supported - not yet implemented
    $endif

' Configure Serial io                                                                                               
    $ifdef OSENABLE_SERIO_INTERRUPT        ' Introduces OS_WaitRXData
      Declare Hserial_Clear True           ' Clear the overflow bit before reading
      $ifndef OSENABLE_INTERRUPTS          ' Must have interrupts enabled
        $error "OSENABLE_INTERRUPTS required for SERIO Interrupt"
      $endif
      $ifndef OSENABLE_EVENTS              ' and events enabled for Ser I/O
        $error "OSENABLE_EVENTS required for SERIO Interrupt"
      $endif

      $defeval OSEVENTS_COUNT $eval OSEVENTS_COUNT + 3 ' Extends the events count by 3 event ECB

      $if OSENABLE_BUFFERED_SERIO = True 
        $ifndef cUSART1_RxBufferSize 
          $define cUSART1_RxBufferSize 32    ' Set up the buffer size for input 
        $endif
        $ifndef cUSART1_TxBufferSize
          $define cUSART1_TxBufferSize 64    ' and output circular buffers 
        $endif
      $else
        $define OSENABLE_RX_INTERRUPT True   ' enable UART Input Interrupt
        $define OSENABLE_TX_INTERRUPT True   ' enable USART Transmit interrupt
      $endif
    $endif
                                                                                      
' Configure Error Handler
 
    $ifdef  OSENABLE_ERROR_HANDLER           ' Introduces ErrorHandler task
      $defeval OSTASKS_COUNT $eval OSTASKS_COUNT + 1 ' so increase Task count by 1
      $ifndef OSWATCHDOG_TIMEOUT
        $define OSWATCHDOG_TIMEOUT 20        ' Tick count before OSWatchdog triggers
      $endif
    $endif

' Public Aliases to task states ---------------------------------------------------
    $define OSTCB_DESTROYED 0       ' Destroyed
    $define OSTCB_TASK_STOPPED 1    ' Task Stopped
    $define OSTCB_TASK_DELAYED 2    ' Delayed n OSticks
    $define OSTCB_TASK_WAITING 3    ' Waiting on an event
    $define OSTCB_TASK_ELIGIBLE 4   ' Ready to run
    $define OSTCB_TASK_RUNNING 5    ' Running

' Public Aliases to Cyclic Timer states -------------------------------------------
    $define OSTCB_CYCTMR_STOPPED 1  ' CycTimer Stopped
    $define OSTCB_CYCTMR_WAITING $B ' Waiting on timeout
    $define OSTCB_CYCTMR_ELIGIBLE $C' Ready to run
    $define OSTCB_CYCTMR_RUNNING $D ' Running

' Public Aliases to Cyclic Timer Modes --------------------------------------------
    $define OSCT_CONTINUOUS $50
    $define OSCT_ONE_SHOT $51
    $define OSCT_CONTINUOUS_DONT_START $1050
    $define OSCT_ONSHOT_DONT_START $1051

' Public Aliases to Event Types ---------------------------------------------------
    $define OSECB_EVENT_NOEVENT 0   ' Destroyed
    $define OSECB_EVENT_BINARY 1    ' Binary Semaphore Event
    $define OSECB_EVENT_COUNTING 2  ' Counting Semaphore Event
    $define OSECB_EVENT_MESSAGE 3   ' Intertask Message Event

' Misc
    $define OSNO_TIMEOUT 0          ' Use in Create Task
    $define OSNO_DELAY 0            ' Use in Create Task or Cyclic timer
    $define OSSTARTED True          ' Use in Create Task or Cyclic timer
    $define OSSTOPPED False         ' Use in Create Task or Cyclic timer
    $define OSNO_TASK $FFFF         ' Task not assigned or create error
    $define OSCURR_TASK $FFFF       ' Used for getting current task details
    $define OSNO_EVENT $FFFF        ' Event not assigned or create error
    $define OSNO_MESSAGE 0          ' Message pointer points to no message
'---------------------------------------------------------------------------------------------------

    Dim CurrentTask As Word  ' TaskID - Ranges from 0 to OSTask_Count
    Dim CurrTaskPtr As Word  ' TaskID * TCBSize  Absolute offsets into Tasks
    Dim CurrRegsPtr As Word  ' Address into RegList (AddressOf(Reglist) + TaskID * RegsSize * 2
    Dim LastTask    As Word  ' TaskID holds the TaskID of the last task created
    Dim LastExecutedTask[OSPRIO_COUNT] As Word Heap
    Dim LowestPriority As Word

    Dim OSPW0 As Word        ' General purpose variables used internally
    Dim OSPW1 As Word
    Dim OSPW2 As Word
    Dim OSPW3 As Word        '
'    Dim OSPW4 As Word
    Dim OSMPUStatus As Word  ' Temporary processor STATE storage
    Dim OSState As Word      ' Reflects the running state of the OS.

    $define TCBSize 8

    Dim Tasks[OSTASKS_COUNT * TCBSize] As Word Heap

    ' TCB Offsets for Tasks
    Symbol _State    = 0
    Symbol _Priority = 1
    Symbol _EventID  = 2
    Symbol _Timer    = 3
    Symbol _Misc     = 4
    Symbol _Dummy    = 5     ' Have to keep PCH, PCL on byte boundary
    Symbol _PCHW     = 6
    Symbol _PCLW     = 7
    ' TCB Offsets for Cyclic Timer
    Symbol _Period   = 2
    Symbol _Counter  = 3
    Symbol _Delay    = 5
    ' Misc bit settings
    Symbol mi_FIRSTTIME     = $1000                     '
    Symbol mi_SET_TIMEOUT   = $0100                     ' Set Timeout bit = OR with MISC
    Symbol mi_CLEAR_TIMEOUT = $FEFF                     ' Clear Timeout Bit = AND with MISC

    $if _type = _PIC24E
      $define REGSize 19
    $else
      $define REGSize 18
    $endif

    Dim RegList[OSTASKS_COUNT * REGSize] As Word Heap       ' Holds the context of each task
    ' W0     = 0
    ' through          - WReg Storage
    ' W14    = 28
    ' RCOUNT = 30      - SFR Storage
    ' TBLPAG = 32
    ' DSRPAG = 34      - PIC24E
    ' DSWPAG = 36      - PIC24E
    ' PSVPAG = 34      - PIC24F

    $ifdef OSENABLE_EVENTS
      $if OSENABLE_MESSAGES = True
        $define ECBSize 4
      $else
        $define ECBSize 3
      $endif

      $define OSECB_BINARY_MESSAGE 1

      Dim LastEvent    As Word   ' EventID - Multiply by ECB Size to get index into Event array
      Dim ECBs[OSEVENTS_COUNT * ECBSize] As Word Heap
  'offsets into ECB
      Symbol _ETYPE    = 0
      Symbol _SEMCOUNT = 1
      Symbol _SEMISR   = 2
      Symbol _MSGPTR_H = 2
      Symbol _MSGPTR_L = 3

      $if OSENABLE_BUFFERED_SERIO = True
        Include  "Buffer_Tx1_Rx1_24.inc"
      $endif

      $if OSENABLE_RX_INTERRUPT = True
        Dim e_RXEvent As Word
        Dim e_RXOErr As Word
      $endif

      $if OSENABLE_TX_INTERRUPT = True
        Dim e_TXEvent As Word
      $endif

      $ifdef OSENABLE_ERROR_HANDLER
        Dim LastError As Word
        Dim t_Error As Word
        Dim OSTmpDW As Dword
        $ifdef OSWATCHDOG_TIMEOUT
          Dim OSWatchdog As Word
        $endif
        Include "P24RTOSErrors.inc"
      $endif

    $endif

;----Timers-----------------------------------------------------------------------------------------
' Holds OS System Timer value
    $if OSENABLE_TIMER = True
      $if OSTICK_CTR_SIZE = 0
      $elseif OSTICK_CTR_SIZE = 2
        Dim OSTickCount As Word
      $elseif OSTICK_CTR_SIZE = 4
        Dim OSTickCount As Dword
      $else
        $error "OSTICK_CTR_SIZE must be 2, or 4 bytes"
      $endif
      $if OSTICK_CTR_EVENT = True
        Dim e_TCTREvent As Word
        $if OSTICK_CTR_SIZE = 2
          Dim OSTickCtrEventVal As Word
        $elseif OSTICK_CTR_SIZE = 4
          Dim OSTickCtrEventVal As Dword
        $endif
        $define OSTICK_CTR_EVENT e_TCTREvent
      $endif
    $endif
    
    GoTo OverPRTOS     'Skip over subroutines
    

'===================================================================================================
'Interrupt setup routine and ISR, call this proc from Init.

    $ifdef OSENABLE_INTERRUPTS
      Proc OSInterrupts_Init()
        Dim tmpW As WREG4
        Dim TskCnt As WREG5          '

        $if OSTICK_SOURCE = T1
          TMR1 = 0
          PR1 =  OSTIMER_PRELOAD     ' Load Timer1 period
          T1CON = %0010000000010000  ' Stop Timer1, Discontinue operation in Idle mode
                                     ' Gated time accumulation disabled,  Internal clock
                                     ' Do not synchronise external clock input
          $if OSTIMER_PRESCALE = 256 Or OSTIMER_PRESCALE = 64 Or OSTIMER_PRESCALE = 8  Or OSTIMER_PRESCALE = 1
            $if OSTIMER_PRESCALE <= 8
              T1CONbits_TCKPS1 = 0
              $if OSTIMER_PRESCALE = 8
                T1CONBits_TCKPS0 = 1    ' Prescaler 8:1
              $else
                T1CONBIts_TCKPS0 = 0    ' Prescaller 1:1
              $endif
            $else
              T1CONbits_TCKPS1 = 1
              $if OSTIMER_PRESCALE = 256
                T1CONBIts_TCKPS0 = 1    ' Prescaler 256:1
              $else
                T1CONBIts_TCKPS0 = 0    ' Prescaler 64:1
              $endif
            $endif
          $else
            $error "OSTIMER Prescale value must be 256, 64, 8 ot 1"
          $endif
          IFS0bits_T1IF  = 0            ' Clear Timer1 interrupt flag
          IPC0Bits_T1IP2 = 1
          IPC0BIts_T1IP1 = 1
          IPC0bits_T1IP0 = 1            ' Set priority
          IEC0bits_T1IE  = 1            ' Enable the Timer1 interrupt
        $endif
        $if OSTICK_SOURCE = T2
          TMR2 = 0
          PR2 = OSTIMER_PRELOAD         ' Load Timer2 period
          T2CON = %0010000000000000     ' Stop Timer2, Discontinue operation in Idle mode
                                        ' Gated time accumulation disabled Timer2 as 16-bit timer
                                        ' Internal clock
          $if OSTIMER_PRESCALE = 256 Or OSTIMER_PRESCALE = 64 Or OSTIMER_PRESCALE = 8  Or OSTIMER_PRESCALE = 1
            $if OSTIMER_PRESCALE <= 8
              T2CONbits_TCKPS1 = 0
              $if OSTIMER_PRESCALE = 8
                T2CONBits_TCKPS0 = 1    ' Prescaler 8:1
              $else
                T2CONBIts_TCKPS0 = 0    ' Prescaller 1:1
              $endif
            $else
              T2CONbits_TCKPS1 = 1
              $if OSTIMER_PRESCALE = 256
                T2CONBIts_TCKPS0 = 1    ' Prescaler 256:1
              $else
                T2CONBIts_TCKPS0 = 0    ' Prescaler 64:1
              $endif
            $endif
          $else
            $error "OSTIMER Prescale value must be 256, 64, 8 ot 1"
          $endif
          IFS0bits_T2IF  = 0            ' Clear Timer2 interrupt flag
          IPC1bits_T2IP0 = 0            ' Set priority
          IEC0bits_T2IE  = 1            ' Enable the Timer2 interrupt
        $endif
        $if OSTICK_SOURCE = EXT
          GoSub OSTICK_EXT_INIT         ' User defined pin and interrupt source
        $endif

        $ifdef OSWATCHDOG_TIMEOUT
          OSWatchdog = OSWATCHDOG_TIMEOUT    ' Prime Watchdog timer.
        $endif

        $if OSENABLE_RX_INTERRUPT = True
          U1MODEBits_UARTEN = 1              ' UART Enabled
          U1MODEBits_USIDL = 1               ' Discontinue when device in idle
          U1MODEBits_IREN = 0                ' IrDA encoding Off
          U1MODEBits_UEN1 = 0                ' TX and RX pins enbled
          U1MODEBits_UEN0 = 0                ' FLow control pins off
          U1MODEBits_WAKE = 0                ' No Wake on start bit
          U1MODEBits_LPBack = 0              ' Loopback off
          U1MODEBits_ABAUD = 0               ' Auto baud detect Off
          U1MODEBits_RXINV = 0               ' Idle State is high
          U1MODEBits_BRGH = 0                ' 16 clocks per bit
          U1MODEBits_PDSEL1 = 0              ' 8 data bit
          U1MODEBits_PDSEL0 = 0              ' no parity
          U1MODEBits_STSEL = 0               ' One Stop Bit

          U1STABits_URXISEL1 = 0             ' Interrupt on character received
          U1STABits_URXISEL0 = 0             ' and in Receive buffer
          IEC0Bits_U1RXIE = 1                ' Enable RX Interrupt
          e_RXEvent = OSCreateSem(0)         ' Create Data Received Event
          e_RXOErr  = OSCreateBinSem(False)  ' Create Overflow Error Event
        $endif

        $if OSENABLE_TX_INTERRUPT = True
          U1STABits_UTXISEL1 = 0             ' Interrupt when data transferred
          U1STABits_UTXISEL0 = 0             ' from TX Buffer to shift register
          U1StaBits_UTXEN = 1                ' Enable TX
          IEC0Bits_U1TXIE = 0                ' Disable TX Interrupt - will be enabled by OSRUN()
          e_TXEvent = OSCreateBinSem(False)  ' Create Data Transmitted event
        $endif

'---------------------------------------------------------------------------------------------------
' OS_Timer Interrupt Service Routine

' Sets a tick flag which is cleared by scheduler and increments the OSTickCount
        $if OSTICK_SOURCE = T1 
          $define ClrTmrflg IFS0bits_T1IF  = 0        ' Clear Timer1 interrupt flag
          $define EnTmrIntr IEC0bits_T1IE  = 1        ' Enable the Timer1 interrupt
          Isr T1Interrupt, UnHandled
        $endif
        $if OSTICK_SOURCE = T2 
          $define ClrTmrflg IFS0bits_T2IF  = 0        ' Clear Timer2 interrupt flag
          $define EnTmrIntr IEC0bits_T2IE  = 1        ' Enable the Timer2 interrupt
          Isr T2Interrupt UnHandled
        $endif

        Dim TaskPtr  As OSPW1
        Dim TmpWrd   As WREG4
        Dim NextTask As WREG5
        
        'Save Context
        Disi #0x3fff
        Push STATUS
        Push RCOUNT
        Push CORCON
        Push.w W14
        Mov.w #0x00,W14
        mRepeat #13
        Mov.w [W14++],[W15++]
        $if _type = "_PIC24E" Or _type = "_PIC33E"
        $define _OFFSET 48
        Push DSRPAG
        Push DSWPAG
        $else
        $define _OFFSET 46
        Push PSVPAG
        $endif
        Push PBP_VAR0
        Push PBP_VAR1
        Disi #0x01
        
        High PORTB.8                                  ' debug to Virtual Scope

        ClrTmrflg                                     ' Clear Timer interrupt flag
        $if OSTICK_CTR_SIZE > 0
          Inc OSTickCount                             ' increment the tick count
          $if OSTICK_CTR_EVENT = True
            If OSTickCount = OSTickCtrEventVal Then   ' if Tick Count reached event value
              OSSignalSemaphore_ISR(e_TCTREvent)      ' Signal Tick Counter event
            EndIf
          $endif
        $endif

        $if OSENABLE_TIMEOUTS = True
          If LastTask <> OSNO_TASK Then
            TaskPtr = 0
            NextTask = 0
            Do                                ' run through all tasks for events and timers/Timeouts
            If Tasks[TaskPtr] <> OSNO_TASK Then       ' Skip any Unused slots
              If Tasks[TaskPtr] & $2 = $2 Then        ' If Task Delayed, Waiting or CycTmr Running
                TmpWrd = Tasks[TaskPtr + _Timer]      ' Get the timer
                If TmpWrd > 0 Then                    ' Timer decrements to 0
                  Dec TmpWrd                          ' Still +ve so decrement it
                  Tasks[TaskPtr + _Timer] = TmpWrd    ' and save it back in the TCB
                  If TmpWrd = 0 Then                  ' If it has reached 0 time to act
                    Select Tasks[TaskPtr]
                    Case OSTCB_TASK_DELAYED           ' if it was delayed task it is time to
                      Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE ' make it eligible to run again
                      $ifdef OSENABLE_EVENTS
                      Case OSTCB_TASK_WAITING         ' if waiting task then has timed out
                        Tasks[TaskPtr + _Misc] = Tasks[TaskPtr + _Misc] | mi_SET_TIMEOUT' set timeout flag
                        Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE ' and make eligible to run again.
                      $endif
                      $if OSENABLE_CYCLIC_TIMERS = True
                      Case OSTCB_CYCTMR_WAITING
                        Tasks[TaskPtr] = OSTCB_CYCTMR_ELIGIBLE ' Update the timer
                      $endif
                    EndSelect
                  EndIf
                EndIf
              EndIf
            EndIf
            Inc NextTask
            TaskPtr = NextTask * TCBSize              ' Step on to next task
            Loop Until NextTask > LastTask            ' Loop through all tasks
          EndIf
        $endif
        $ifdef OSWATCHDOG_TIMEOUT
          If OSWatchdog > 0 Then 
            Dec OSWatchdog                            ' Decrement the watchdog each OS Tick
          Else                                        ' if Watchdog has timed out
            LastError.HighByte = OSSCHEDULER_TIMEOUT  ' Set up the error code and 
            LastError.LowByte = CurrentTask.LowByte   ' the task ID of the timed out task
            Tasks[CurrTaskPtr] = OSTCB_TASK_STOPPED   ' Stop hung task
            ' locate return address in the stack and replace it with Scheduler address
            TaskPtr = t_Error * TCBSize               ' Point to the Error Task and
            Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE      ' set it to run on next Select_Run_Task
            OSTmpDW = AddressOf(Select_Run_Task)      ' Get the address of the scheduler and 
            WREG14 = WREG15                           ' Load WREG14 with the stack Pointer            
            WREG14 = WREG14 - _OFFSET                 ' Adjust pointer to LSWord of the ISR Return
            WREG13 = OSTmpDW.LowWord                  ' Get Low Word of the error routine and
            Mov.w W13, [W14++]                        ' place in ISR return value low word
            WREG13 = OSTmpDW.HighWord                 ' Get High Word of error routine
            Mov.w W13, [W14++]                        ' place in ISR Return value high word
          EndIf
        $endif
        EnTmrIntr                                     ' enable timer Interrupt

        Low PORTB.8                                   ' debug to Virtual Scope      

        ' Restore Context - This equates to 44/46 decrements of the stack
        Disi #0x3fff
        Pop PBP_VAR1
        Pop PBP_VAR0
        $if _type = "_PIC24E" Or _type = "_PIC33E"
        Pop DSWPAG
        Pop DSRPAG
        $else
        Pop PSVPAG
        $endif
        Mov.w #0x1a,W14
        mRepeat #13
        Mov.w [--W15],[W14--]
        Pop.w W14
        Pop CORCON
        Pop RCOUNT
        Pop STATUS
        Disi #0x01
        EndIsr        ' and return

'---------------------------------------------------------------------------------------------------
' OS_Serial Interrupt Service Routines

        ' Signals a Data Receive event and/or a Receive Error event
        $if OSENABLE_RX_INTERRUPT = True
          Isr U1RXInterrupt
            If U1STA & $000E > 0 Then                           ' Check for receive errors
              OSSignalSemaphore_ISR(e_RXOErr)                   ' Raise error event
              U1STABits_OERR = 0                                ' Clear Possible Overrun error
            EndIf
            If U1STaBits_URXDA = 1 Then
              OSSignalSemaphore_ISR(e_RXEvent)                  ' Signal data available
            EndIf
            Clear IFS0bits_U1RXIF                               ' Clear the RX Interrupt flag
          EndIsr
        $endif

        ' Signals when the data buffer becomes empty
        $if OSENABLE_TX_INTERRUPT = True
          Isr U1TXInterrupt
            Clear IFS0bits_U1TXIF                               ' Clear the TX interrupt flag
            OSSignalSemaphore_ISR(e_TXEvent)                    ' Signal Data Sent event
          EndIsr
        $endif

      EndProc
    $endif

'=============================== Internal Routines - Inline ========================================

    $if OSTICK_SOURCE = T1
      $define Disable_OS_Timer() Clear T1CONbits_TON ' Stop Timer
      $define Enable_OS_Timer()  Set T1CONbits_TON ' Start Timer
    $else
      $define Disable_OS_Timer() Clear T2CONBits_TON
      $define Enable_OS_Timer()  Set T2CONBits_TON
    $endif

    $define Disable_Interrupts() OSMPUStatus = Sr  '
    Sr = Sr | $E0
    $define Enable_Interrupts()  Sr = OSMPUStatus

'$define GetReturnAddress()  Disable_Interrupts()                                          '
'                            OSPW0 = WREG15             ; Save Stack pointer               '
'                            OSPW3 = WREG14             ; Save WREG14                      '
'                            OSPW2 = WREG13             ; and WREG13                       '
'                            OSPW1 = CurrTaskPtr + _PCHW ; TCB + PCL offsets into Tasks    '
'                            OSPW1 = OSPW1 << 1         ; Double as address in bytes       '
'                            WREG13 = AddressOf(Tasks) + OSPW1                             '
'                            WREG14 = OSPW0             ; place Stack pointer in WREG14    '
'                            Dec WREG14                 ; Adjust stack pointer back     '
'                            Dec WREG14                 ; up to Return address             '
'                            Mov [W14--], [W13++]       ; Save PCH in TCB for Current task '
'                            Mov [W14], [W13]           ; Save PCL in TCB                  '
'                            WREG13 = OSPW2             ; Restore WREG13                   '
'                            WREG14 = OSPW3             ; and WREG14                       '
'                            GoSub Save_Context         ; Save W0 through W14              '
'                            Enable_Interrupts()

    $define GetReturnAddress()  Disable_Interrupts()                 '
    OSPW0 = WREG15                       '
    OSPW3 = WREG14                       '
    OSPW2 = WREG13                       '
    OSPW1 = CurrTaskPtr + _PCHW          '
    OSPW1 = OSPW1 << 1                   '
    WREG13 = AddressOf(Tasks)            '
    WREG13 = WREG13 + OSPW1              '
    WREG14 = OSPW0                       '
    Dec WREG14                           '
    Dec WREG14                           '
    Mov [W14--], [W13++]                 '
    Mov [W14], [W13]                     '
    WREG13 = OSPW2                       '
    WREG14 = OSPW3                       '
    GoSub Save_Context                   '
    Enable_Interrupts()

' ===================================Internal Routines==============================================

' The scheduler - Selects next task, Restores Registes for task and transfers control to new task.
' All task switching procedures actually return through here ---------------------------------------

    $define SelectAndRunTask GoTo Select_Run_Task

    Select_Run_Task:
    Do
    Select_Next_Task()         ; loop until a waiting task is found
    $ifdef OSWATCHDOG_TIMEOUT
      OSWatchdog = OSWATCHDOG_TIMEOUT  ; set the watchdog timeout
    $endif
    Loop Until CurrentTask <> OSNO_TASK;
    Disable_Interrupts()         ; waiting task found set task state
    If Tasks[CurrTaskPtr] = OSTCB_TASK_ELIGIBLE Then
      GoSub Restore_Context      ; Restore all 14 registers for tasks only
      Tasks[CurrTaskPtr] = OSTCB_CYCTMR_RUNNING
    Else
      Tasks[CurrTaskPtr] = OSTCB_TASK_RUNNING
    EndIf
    OSPW3 = WREG14               ; Save WREG14
    OSPW2 = WREG13               ; and WREG13
    WREG14 = WREG15              ; Stack pointer to WREG14
    Dec WREG14                   ; Adjust stack pointer to
    Dec WREG14                   ; start of return address
    OSPW1 = CurrTaskPtr + _PCHW  ; Relative pos of PCH Tasks inTasks
    OSPW1 = OSPW1 << 1           ; Adjust for Byte address
    WREG13 = AddressOf(Tasks) + OSPW1 ; Address of PCH in WREG13
    Mov [W13++], [W14--]         ; place PCH and PCL for next task
    Mov [W13], [W14]             ; onto Stack ready for next task
    WREG14 = OSPW3               ; Restore WREG14
    WREG13 = OSPW2               ; Restore WREG13
    Enable_Interrupts()
    Return

'---------------------------------------------------------------------------------------------------
' Check Events - checks ISR signalled events, then checks all waiting tasks aginst signalled events
' and changes the task state of anmy matching task to Avaliable

    $ifdef OSENABLE_EVENTS | OSENABLE_TIMER = True

      Proc CheckEvents()
        Dim EventPtr  As OSPW0
        Dim TaskPtr   As OSPW1
        Dim NextEvent As WREG3
        Dim NextTask  As WREG5
        Dim TmpWrd    As WREG4                          ' temp storage for SemCount

        Push   WREG3                                    ' we are using WREGs 3,4,5 so save them
        Push.d WREG4
  ' Check if events were signalled from ISR
        $ifdef OSENABLE_EVENTS
          Disable_Interrupts
          If LastEvent <> OSNO_EVENT Then
            EventPtr = 0
            NextEvent = 0
            Do                                          ' Run through all the events
            If ECBs[EventPtr] <> OSECB_EVENT_NOEVENT Then
              If ECBs[EventPtr + _SEMISR] > 0 Then      ' Event Signalled from an ISR
                If ECBs[EventPtr] = OSECB_EVENT_BINARY Then
                  TmpWrd = 1                            ' Binary so Set SemCount to 1 (Signalled)
                Else
                  TmpWrd = ECBs[EventPtr + _SEMCOUNT]   ' Get the SemCount value
                  TmpWrd = TmpWrd + ECBs[EventPtr + _SEMISR]' Add SemISR to SemCount
                EndIf
                Clear ECBs[EventPtr +_SEMISR]           ' Clear SemISR down as now added to SemCount
                ECBs[EventPtr + _SEMCOUNT] = TmpWrd     ' Save it back to SemCount
              EndIf
            EndIf
            Inc NextEvent
            EventPtr = NextEvent * ECBSize
            Loop Until NextEvent > LastEvent            ' Loop through all events
          EndIf
          Enable_Interrupts                             ' Enable Interrupts
        $endif
  ' Check all waiting tasks
        If LastTask <> OSNO_TASK Then
          TaskPtr = 0
          NextTask = 0
          Do
          If Tasks[TaskPtr] <> OSNO_TASK Then           ' check for waiting tasks and make eligible
            If Tasks[TaskPtr] = OSTCB_TASK_WAITING Then ' if Task is waiting
              EventPtr = Tasks[TaskPtr + _EventID]      ' Get the Event
              EventPtr = EventPtr * ECBSize             ' ECB its waiting for
              If ECBs[EventPtr + _SEMCOUNT] > 0 Then    ' Check SemCount and
                Dec ECBs[EventPtr + _SEMCOUNT]          ' Decrement if not zero
                Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE    ' Set Task to Eligible
              EndIf
            EndIf
          EndIf
          Inc NextTask                                  ' Step on to next task
          TaskPtr = NextTask * TCBSize
          Loop Until NextTask > LastTask                ' Loop through all tasks
        EndIf
        Pop.d WREG4
        Pop   WREG3
      EndProc
    $endif

    Proc Select_Next_Task()
      Dim TaskPtr  As WREG4
      Dim NextTask As WREG5
      Dim Priority As OSPW0

      Push.d WREG4
      CurrentTask = OSNO_TASK                       ' clear Current task
      If LastTask <> OSNO_TASK Then
        $ifdef OSENABLE_EVENTS
          CheckEvents()                             ' Update any TCBs for events
        $endif
        Priority = 0
        Do
        If LastExecutedTask[Priority] = OSNO_TASK Then LastExecutedTask[Priority] = LastTask
        NextTask = LastExecutedTask[Priority]       ' Round robin at each priority level
        Do
        Inc NextTask
        If NextTask > LastTask Then NextTask = 0    ' wrap round
        TaskPtr = NextTask * TCBSize                ' TaskPtr holds offset into Tasks for current task
        If Priority = Tasks[TaskPtr + _Priority] Then ' if Tasks Priority = current priority
          If Tasks[TaskPtr] & 4 = OSTCB_TASK_ELIGIBLE Then ' and task is Eligible
            LastExecutedTask[Priority] = NextTask   ' We have found highest priority eligible task
            CurrentTask = NextTask                  ' TaskID of Current Task
            CurrTaskPtr = CurrentTask * TCBSize     ' Set offset for Current Task into Tasks
            CurrRegsPtr = CurrentTask * REGSize     ' and offset for Current Task into Regs
            GoTo Done_Next_Task
          EndIf
        EndIf
        Loop Until NextTask = LastExecutedTask[Priority]
        Inc Priority                                ' on to next priority level
        Loop Until Priority > LowestPriority
      EndIf
      Done_Next_Task:
      Pop.d WREG4
    EndProc

' Restores the context of the task about to run (Current_Task) from the RegList using the offset
' into the RegList for the current task held in CurrRegsPtr

    Restore_Context:
    Push RegList[CurrRegsPtr]                 ' Save restore values of WREG14 and WREG13 to stack
    Push RegList[CurrRegsPtr + 2]
    WREG14 = AddressOf(RegList) + CurrRegsPtr ' Point WREG14 at entry in RegList for current Task
    WREG13 = $00
    mRepeat #12
    Mov.w [W14++],[W13++]           ' Restore Wregs 0-12 from RegList
    WREG14 = WREG14 + 4             ' Skip past space for WREG13 and 14
    ' Now load a few other SFRs
    Mov [W14++], W13
    Mov W13, RCOUNT                 ; Restore RCOUNT
    Mov [W14++], W13
    Mov W13, TBLPAG                 ; Restore TBLPAG
    $if _type = _PIC24E
      Mov [W14++], W13
      Mov W13, DSRPAG               ; Restore DSRPAG
      Mov [W14++], W13
      Mov W13 DSWPAG                ; Restore DSWPAG
    $else
      Mov [W14], W13
      Mov W13, PSVPAG               ; Restore PSVPAG
    $endif
    Pop WREG13                      ; Finally Restore WREG13 & WREG14 from Stack
    Pop WREG14
    Return

' Save the context of the current task which is about to yield to the Reglist using the offset into
' the reglist for the current task held in CurrRegPtr

    Save_Context:
    Push WREG13
    Push WREG14
    WREG14 = AddressOf(RegList) + CurrRegsPtr
    WREG13 = $00
    mRepeat #12
    Mov.w [W13++],[W14++]           ; Save WRegs 0-12 to RegList
    WREG14 = WREG14 + 4             ; Skip past space for WREG13 and WREG14
    ' Now save some FSRs
    Mov [W14++], W13
    Mov W13, RCOUNT                 ; Save RCOUNT
    Mov [W14++] , W13
    Mov W13, TBLPAG                 ; Save TBLPAG
    $if _type = _PIC24E
      Mov [W14++], W13
      Mov W13, DSRPAG               ; Save DSRPAG
      Mov [W14++], W13
      Mov W13 DSWPAG                ; Save DSWPAG
    $else
      Mov [W14], W13
      Mov W13, PSVPAG               ; Save PSVPAG
    $endif
    Pop RegList[CurrRegsPtr + 28] ' Get saved WREG14 from Stack and save in RegList
    Pop RegList[CurrRegsPtr + 24] ' Get saved WREG13 from stack and save in RegList
    Return

 '--------------------------------------------------------------------------------------------------
 ' Helper routines

 ' Loads a TCB with task details - called by CreateTask or ReplaceTask
    Proc LoadTask(TaskID As Word, TaskAddr As Dword, Priority As Word, Start As Word)
      Dim TaskPtr As OSPW0
      If Priority >= OSPRIO_COUNT Then Priority = OSPRIO_COUNT - 1;
      If Priority > LowestPriority Then LowestPriority = Priority;
      TaskPtr =  TaskID * TCBSize                     ' Get index into task array
      If Start IsTrue Then
        Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE          ' Set State to Eligible
      Else
        Tasks[TaskPtr] = OSTCB_TASK_STOPPED           ' Set State to Stopped
      EndIf
      Tasks[TaskPtr + _Priority] = Priority           ' Set Priority
      Tasks[TaskPtr + _EventID] = OSNO_EVENT          ' clear eventID
      Tasks[TaskPtr + _Timer] = 0                     ' Clear Timer value
      Tasks[TaskPtr + _Misc] = mi_FIRSTTIME           ' Set Misc Field
      Tasks[TaskPtr + _PCHW] = TaskAddr.HighWord      ' Set the Starting address of the Task
      Tasks[TaskPtr + _PCLW] = TaskAddr.LowWord       ' in the TCB - will change when Task yields
    EndProc

'---------------------------------------------------------------------------------------------------
' Returns a new TaskID for the first vacant TCB found

    Proc GetTaskID(), Word                            ' Returns the first vacant TCB TaskID
      Dim TaskPtr As OSPW0
      Dim TaskID As OSPW1
      TaskPtr = 0
      TaskID = 0
      If LastTask = OSNO_TASK Then                    ' If no task yet assigned
        LastTask = TaskID                             ' Set LastTask to 0
        Result = TaskID                               ' and return with TaskID in result
      Else
        Result = OSNO_TASK                            ' default to No task
        Do                                            ' scan through the TCBs for empty TCB
        If Tasks[TaskPtr] = OSTCB_DESTROYED Then      ' if empty slot found
          If TaskID > LastTask Then LastTask = TaskID ' Update lastTask
          Result = TaskID                             ' set return with taskID of TCB
          Break                                       ' and break as we have finished
        Else
          Inc TaskID                                  ' TCB occupied so inc TaskID and step
          TaskPtr = TaskPtr + TCBSize                 ' on to next TCB
        EndIf
        Loop Until TaskID >= OSTASKS_COUNT            ' loop until end of Tasks List
      EndIf
    EndProc                                           ' returns with either OSNO_TASK or new TaskID

'---------------------------------------------------------------------------------------------------
' Returns a new EventID for the first vacant ECB found

    $ifdef OSENABLE_EVENTS
      Proc GetEventID(), Word                         ' Returns the first vacant ECB EventID
        Dim EventPtr As OSPW0
        Dim EventID As OSPW1
        EventPtr = 0
        EventID = 0
        If LastEvent = OSNO_EVENT Then                ' If no Event yet assigned
          LastEvent = EventID                         ' Set last event to 0
          Result = LastEvent                          ' and return with EventID in result
        Else
          Result = OSNO_EVENT                         ' defualt to No Event
          Do                                          ' scan through the TCBs for empty TCB
          If ECBs[EventPtr] = OSECB_EVENT_NOEVENT Then' if empty slot found
            If EventID > LastEvent Then LastEvent = EventID' Update lastEvent
            Result = EventID                          ' set return with taskID of TCB
            Break                                     ' and break as we have finished
          Else
            Inc EventID                               ' ECB occupied so inc EventID and step
            EventPtr = EventPtr + ECBSize             ' on to next ECB
          EndIf
          Loop Until EventID >= OSEVENTS_COUNT        ' loop until end of Tasks List
        EndIf
      EndProc                                         ' returns with either OSNO_EVENT or new EventID
    $endif

'==================================== Published Routines ===========================================

    Proc OSVerNo(), Dword                             ' Returns Version number of OS
      Result.Byte3 = OSMajorVer                       
      Result.Byte2 = OSMinorVer
      Result.Byte1 = OSRelease
      Result.Byte0 = oSBuild
    EndProc
    
    Proc OSInit()
      Dim Index As OSPW0
      LowestPriority = 0                              ' Initialise a few variables
      LastTask = OSNO_TASK
      CurrentTask = OSNO_TASK
      CurrTaskPtr = 0
      CurrRegsPtr = 0
      OSState = 0      
      Clear Tasks                                     ' Clear Tasks to Destroyed
      $ifdef OSENABLE_EVENTS
        Clear ECBs                                    ' Clear Events to Destroyed
        LastEvent = OSNO_EVENT
        $ifdef OSENABLE_ERROR_HANDLER
          OSErrorHdlr_Init()                          ' Initialise the Error handler
        $endif
      $endif
      Set LastExecutedTask                            ' set LastExecutedTask Array to OSNO_TASK
      $ifdef OSENABLE_INTERRUPTS
        OSInterrupts_Init()                           ' enables timer and usart interrupts
        $if OSENABLE_BUFFERED_SERIO = True
          USART1_Init()                               '  Initialise the USART
        $endif
      $endif

    EndProc

' Starts the P-RTOS24 scheduler.
' Should be the last routine executed in the main program, the task is blocking (never left).

    Proc OSRun()
      $if OSENABLE_TX_INTERRUPT = True
        Set IEC0bits_U1TXIE                           ' Enable TX Interrupt
      $endif
      OSState = True
      Enable_OS_Timer()
      SelectAndRunTask()
    EndProc

'-----------------------
' Task Related Routines
'-----------------------

'---------------------------------------------------------------------------------------------------
' Creates a task - TaskAddr is the address(Label) of the task and priority is the priority level
' Returns with the Index (TaskID) of the entry into the Task Control block or OSNO_TASK if it fails.

    Proc OSCreateTask(ByRef TaskAddr As Dword, Priority As Word, Start As Word), Word
      Result = GetTaskID()                            ' Get new TaskID
      If Result <> OSNO_TASK Then
        LoadTask(Result, TaskAddr, Priority, Start)   ' Load up TCB
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSCREATETASK_ERROR, LIST_FULL)
        $endif
      EndIf
    EndProc

'---------------------------------------------------------------------------------------------------
' Destroys a task, sets TCB slot to empty and sets TaskID to OSNO_TASK if successful
' Task has to be stopped before it is destroyed.

    Proc OSDestroyTask(ByRef TaskID As Word)
      Dim TaskPtr As OSPW0
      Dim TaskIdx As OSPW1
      TaskIdx = Ptr16(TaskID)                         ' Get TaskID
      If TaskIdx > LastTask Then
        $ifdef OSENABLE_ERROR_HANDLER
          OS_RaiseError(OSDESTROYTASK_ERROR, INVALID_TASKID)
        $endif
        Return                                        ' Only destroy tasks which exist
      EndIf
      TaskPtr = TaskIdx * TCBSize                     ' Point to the Task in TCB
      If Tasks[TaskPtr] = OSTCB_TASK_RUNNING Then
        $ifdef OSENABLE_ERROR_HANDLER
          OS_RaiseError(OSDESTROYTASK_ERROR, TASK_RUNNING)
        $endif
        Return ' Cannot destroy a running task
      EndIf
      Tasks[TaskPtr] = OSTCB_DESTROYED                ' and set it to empty
      Ptr16(TaskID) = OSNO_TASK                       ' and set TaskID to NO_TASK
    EndProc

'---------------------------------------------------------------------------------------------------
' Replaces the task in TaskID with a different task

    Proc OSReplaceTask(TaskID As Word, ByRef TaskAddr As Dword, Priority As Word, Start As Word), Word
      Result = OSNO_TASK
      If TaskID > LastTask Then
        $ifdef OSENABLE_ERROR_HANDLER
          OS_RaiseError(OSREPLACETASK_ERROR, INVALID_TASKID)
        $endif
        Return              '
      EndIf
      If TaskID = CurrentTask Then
        $ifdef OSENABLE_ERROR_HANDLER
          OS_RaiseError(OSREPLACETASK_ERROR, OSTASK_RUNNING)
        $endif
        Return           ' task running so leave alone and return with error
      EndIf
      LoadTask(TaskID, TaskAddr, Priority, Start)   ' Load TCB with new task details
    EndProc

'---------------------------------------------------------------------------------------------------
' Starts a Stopped task. pTaskID defines the task and pDelay is the waiting time in ticks before
' task will be started.  If Delay is 0 the task will start as soon as possible.

    Proc OSStartTask(TaskID As Word, Delay As Word)
      Dim TaskPtr As OSPW0
      If TaskID > LastTask Then
        $ifdef OSENABLE_ERROR_HANDLER
          OS_RaiseError(OSSTARTTASK_ERROR, INVALID_TASKID)
        $endif
        Return              ' Only do anything if valid task ID
      EndIf
      TaskPtr =  TaskID * TCBSize                   ' Get index into Tasks Array STATE
      If Tasks[TaskPtr] = OSTCB_TASK_STOPPED Then   ' Only do anything if task curently stopped
        If Delay > 0 Then
          Tasks[TaskPtr] = OSTCB_TASK_DELAYED       ' Set up delayed task
        Else
          Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE      ' Set up imme diate task
        EndIf
        Tasks[TaskPtr + _Timer] = Delay             ' Set the delay value
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSSTARTTASK_ERROR, TASK_RUNNING)
        $endif

      EndIf
    EndProc

'---------------------------------------------------------------------------------------------------
' Stops a task, the task can only be restarted with "OS_StartTask" which will resume the task at the
' point where it was when stopped.  If the task was running it will continue until it yields but will
' not restart until OS_StartTask is called

    Proc OSStopTask(TaskID As Word)
      Dim TaskPtr As OSPW0
      If TaskID <= LastTask Then
        TaskPtr = TaskID * TCBSize                    ' Get index into Tasks Array  STATE
        If Tasks[TaskPtr] > OSTCB_TASK_STOPPED Then
          Tasks[TaskPtr] = OSTCB_TASK_STOPPED
        EndIf
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSSTOPTASK_ERROR, INVALID_TASKID)
        $endif
      EndIf
    EndProc

'---------------------------------------------------------------------------------------------------
' Returns the task ID of the task currently running

    Proc OSCurrentTask(), Word
      Result = CurrentTask                          ' Return the current TaskID
    EndProc

'---------------------------------------------------------------------------------------------------
' Returns the state of task in TaskID

    Proc OSTaskState(TaskID As Word), Word
      Dim TaskPtr As OSPW0
      Result = OSTCB_DESTROYED
      If TaskID <= LastTask Then
        TaskPtr = TaskID * TCBSize
        Result = Tasks[TaskPtr]
      $ifdef OSENABLE_ERROR_HANDLER
      Else
        OS_RaiseError(OSTASKSTATE_ERROR, INVALID_TASKID)
      $endif
      EndIf
    EndProc

'---------------------------------------------------------------------------------------------------
' Returns True is Task defined by taskID is running

    Proc OSTaskRunning(TaskID As Word), Word
      Result = False
      If OSTaskState(TaskID) > OSTCB_TASK_STOPPED Then Result = True
    EndProc

'---------------------------------------------------------------------------------------------------
' Changes the Priority of Non running task

    Proc OSSetPriority(TaskID As Word, Priority As Word)   
      Dim TaskPtr As OSPW0
      If TaskID <= LastTask Then
        If TaskID <> CurrentTask Then
          If Priority >= OSPRIO_COUNT Then LowestPriority = OSPRIO_COUNT -1
          If TaskID <= LastTask Then
            TaskPtr = TaskID * TCBSize
            Tasks[TaskPtr + _Priority] = Priority
          EndIf
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSSETPRIORITY_ERROR, TASK_RUNNING)      
        $endif
        EndIf
      $ifdef OSENABLE_ERROR_HANDLER
      Else
        OS_RaiseError(OSSETPRIORITY_ERROR, INVALID_TASKID)            
      $endif
      EndIf
    EndProc

'---------------------------------------------------------------------------------------------------
' Returns the Priority of any task - If TaskID = OSCURR_TASK the current task will be returned

    Proc OSGetPriority(TaskID As Word), Word
      Dim TaskPtr As OSPW0
      TaskPtr = $FFFF
      Result = TaskPtr
      If TaskID <= LastTask Then
        TaskPtr = TaskID * TCBSize
      Else
        If TaskID = OSCURR_TASK Then
          TaskPtr = CurrentTask
        EndIf
      EndIf
      If TaskPtr <> $FFFF Then
        Result = Tasks[TaskPtr + _Priority]
      EndIf
    EndProc

'---------------------------------------------------------------------------------------------------
' Reset Task - resets the task pointers to start at the initial conditions - task must be stopped

    Proc OSResetTask(TaskID As Word, ByRef TaskAddr As Dword)
      Dim TaskPtr As OSPW0
      If TaskID <= LastTask Then
        If TaskID <> CurrentTask Then 
          TaskPtr = TaskID * TCBSize
          Tasks[TaskPtr + _EventID] = OSNO_EVENT        ' clear eventID
          Tasks[TaskPtr + _Timer] = 0                   ' Clear Timer value
          Tasks[TaskPtr + _Misc] = mi_FIRSTTIME         ' Set Misc Field
          Tasks[TaskPtr + _PCHW] = TaskAddr.HighWord    ' Set the Starting address of the Task
          Tasks[TaskPtr + _PCLW] = TaskAddr.LowWord     ' in the TCB - will change when Task yields
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSRESETTASK_ERROR, TASK_RUNNING)              
        $endif
        EndIf
      $ifdef OSENABLE_ERROR_HANDLER
      Else
        OS_RaiseError(OSRESETTASK_ERROR, INVALID_TASKID)      
      $endif
      EndIf
    EndProc

' -------------------------------------------------------------------------------------------
' Basic Task Switching Routines - (See also OS_WaitSemaphore and OS_WaitMessage) under EVENTS
'--------------------------------------------------------------------------------------------

' Return to the scheduler unconditionally, if no other task eligible scheduler will resume current
' task from where it yielded

    Proc OS_Yield()
      GetReturnAddress()                                ' and return to the scheduler
      If Tasks[CurrTaskPtr] <> OSTCB_TASK_STOPPED Then  ' If task not stopped by another task
        Tasks[CurrTaskPtr] = OSTCB_TASK_ELIGIBLE        ' make it eligible to run again
      EndIf
      SelectAndRunTask()
    EndProc

'---------------------------------------------------------------------------------------------------
' Suspends the current task and returns to the scheduler. The latter will resume the task
' after "Ticks" time

    Proc OS_Delay(Ticks As Word)
      GetReturnAddress()
      Tasks[CurrTaskPtr] = OSTCB_TASK_DELAYED           ' change state to Delayed and
      Tasks[CurrTaskPtr + _Timer] = Ticks               ' Set timer to delay value
      SelectAndRunTask()
    EndProc

'---------------------------------------------------------------------------------------------------
' Stops the current task and starts the task with "TaskID"

    Proc OS_ChangeTo(TaskID As Word)
      Dim TaskPtr As OSPW0
      If TaskID > LastTask Or TaskID = CurrentTask Then Return ' not a valid TaskID or nothing to do
      GetReturnAddress()
      Tasks[CurrTaskPtr] = OSTCB_TASK_STOPPED           ' Stop current task
      TaskPtr = TaskID * TCBSize
      If Tasks[TaskPtr] = OSTCB_TASK_STOPPED Then Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE
      SelectAndRunTask()
    EndProc

'---------------------------------------------------------------------------------------------------
' Replaces current task with Task defined by TaskAddr - task runs from start position
' All referrences to previous task are erased.
' New task now occupies the same TCB slot that previous task occupied as Task ID remains same

    Proc OS_ReplaceTask(ByRef TaskAddr As Dword, Priority As Word)
      If Priority >= OSPRIO_COUNT Then Priority = OSPRIO_COUNT - 1;
      Tasks[CurrTaskPtr] = OSTCB_TASK_ELIGIBLE          ' Set State
      Tasks[CurrTaskPtr + _Priority] = Priority         ' Set Priority
      Tasks[CurrTaskPtr + _EventID] = OSNO_EVENT        ' clear eventID
      Tasks[CurrTaskPtr + _Timer] = 0                   ' Clear Timer value
      Tasks[CurrTaskPtr + _Misc] = mi_FIRSTTIME         ' Set Misc Field
      Tasks[CurrTaskPtr + _PCHW] = TaskAddr.HighWord    ' Set the Starting address of the Task
      Tasks[CurrTaskPtr + _PCLW] = TaskAddr.LowWord     ' in the TCB - will change when Task yields
      SelectAndRunTask()
    EndProc

'---------------------------------------------------------------------------------------------------
' Stops the current task and returns to the scheduler' Call from within current task

    Proc OS_Stop()
      GetReturnAddress()
      Tasks[CurrTaskPtr] = OSTCB_TASK_STOPPED
      SelectAndRunTask()
    EndProc

'---------------------------------------------------------------------------------------------------
' Destroys current task or cyclic Timer and returns to scheduler

    Proc OS_Destroy()
      Tasks[CurrTaskPtr] = OSTCB_TASK_STOPPED
      OSDestroyTask(CurrentTask)
      SelectAndRunTask()
    EndProc

' --------------------------------------------------------------------------------------------------
' Changes Priority of current task and yields to OS

    Proc OS_SetPriority(Priority As Word)
      GetReturnAddress()
      If Priority >= OSPRIO_COUNT Then LowestPriority = OSPRIO_COUNT -1
      Tasks[CurrTaskPtr + _Priority] = Priority
      SelectAndRunTask()
    EndProc

'----------------------
' Tick Counter Routines
' ---------------------

' Returns the value of the Tick Counter

    $if OSENABLE_TIMER = True
      $if OSTICK_CTR_SIZE > 2
        Proc OSGetTicks(), Dword
      $else
        Proc OSGetTicks(), Word
      $endif
      Disable_OS_Timer
      Result = OSTickCount
      Enable_OS_Timer
      EndProc

'---------------------------------------------------------------------------------------------------
' Sets the Tick value that will trigger the Tick Event

      $if OSTICK_CTR_EVENT = True
        $if OSTICK_CTR_SIZE > 2
          Proc OSSetTickEvent(EventValue As Dword)
        $else
          Proc OSSetTickEvent(EventValue As Word)
        $endif
        Disable_OS_Timer
        OSTickCtrEventVal = EventVal
        Enable_OS_Timer
        EndProc
      $endif

'---------------------------------------------------------------------------------------------------
' Sets the starting value of the Tick Counter

      $if OSTICK_CTR_SIZE > 0
        $if OSTICK_CTR_SIZE = 2
          Proc OSSetTicks(TickValue As Word)
        $elseif OSTICK_CTR_SIZE = 4
          Proc OSSetTicks(TickValue As Dword)
        $endif
        Disable_OS_Timer
        OSTickCount = TickValue
        Enable_OS_Timer
        EndProc
      $endif
    $endif

'----------------------
' Cyclic Timer Routines
'----------------------
'
' A cyclic timer will call a Subroutine Delay ticks after starting and Period ticks thereafter
' If Mode set to OSCT_CONTINUOUS the timer will repeatedly call Subroutine every Period ticks
' until stopped. With Mode set to OSCT_ONE_SHOT the routine will be called once after Delay ticks
' and Period Ticks. The timer will start automatically on Create unless OSCT_DONT_START is ORed
' with your chosen mode. Typical usage would be to start a Task at regular intervals
' The cyclic timer occupies a Task Control Block location.
' IMPORTANT : Cyclic timer routine is blocking and should therefore be kept short.
'             Cyclic timer routine does not have the context saved or restored.
'             Cyclic Timer routine must end with OS_CTReturn
'             Cyclic Timers must not call OS Commands which cause a context switch.

    $if OSENABLE_CYCLIC_TIMERS = True
      Proc OSCreateCycTmr(ByRef TimerAddress As Dword, Delay As Word, Period As Word, Mode As Word), Word
        Dim TaskPtr As OSPW0
        Result = GetTaskID()
        If Result <> OSNO_TASK Then
          TaskPtr = Result * TCBSize
          Tasks[TaskPtr + _Priority] = 1                   ' Set to top priority to ensure prompt timing
          Tasks[TaskPtr + _Period] = Period                ' Period (Ticks)
          Tasks[TaskPtr + _Delay] = Delay                  ' Initial Delay (Ticks)
          If Delay > 0 Then
            Tasks[TaskPtr + _Counter] = Delay              ' Set counter to initial delay
          Else
            Tasks[TaskPtr + _Counter] = Period             ' Set counter to period if no delay
          EndIf
          If Mode.12 = 1 Then
            Tasks[TaskPtr] = OSTCB_CYCTMR_STOPPED          ' Set timer state to stopped
          Else
            Tasks[TaskPtr] = OSTCB_CYCTMR_RUNNING          ' Set Timer state to running
          EndIf
          Tasks[TaskPtr + _Misc] = Mode & $EFFF            ' Set Timer mode with Don't start bit cleared
          Tasks[TaskPtr + _PCHW] = TimerAddress.HighWord   ' Set Start High Address of Timer routine
          Tasks[TaskPtr + _PCLW] = TimerAddress.LowWord    ' Set Start Low Address of Timer Routine
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSCREATECYCTMR_ERROR, LIST_FULL)
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Stops the Cyclic Timer if it is running

      Proc OSStopCyclicTimer(TimerID As Word)
        Dim TaskPtr As OSPW0
        If TimerID <= LastTask Then
          TaskPtr = TimerID * TCBSize
          If Tasks[TaskPtr] = OSTCB_CYCTMR_WAITING Then
            Tasks[TaskPtr] = OSTCB_CYCTMR_STOPPED
          EndIf
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSCREATECYCTMR_ERROR, INVALID_TIMERID)
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Starts the cyclick timer if it is not already running

      Proc OSStartCyclicTimer(TimerID As Word)
        Dim TaskPtr As OSPW0
        If TimerID <= LastTask Then
          TaskPtr = TimerID * TCBSize
          If Tasks[TaskPtr] = OSTCB_CYCTMR_STOPPED Then
            Tasks[TaskPtr] = OSTCB_CYCTMR_WAITING
          EndIf
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            LastError.HighByte = OSSTARTCYCLICTIMER_ERR
            LastError.LowByte = OSINVALID_TIMERID
            OSERROR_HANDLER()
            OS_RaiseError(OSSTARTCYCLICTIMER_ERROR, INVALID_TIMERID)
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Resets the cyclic timer to its original state

      Proc OSResetCyclicTimer(TimerID As Word)
        Dim TaskPtr As OSPW0
        If TimerID <= LastTask Then
          TaskPtr = TimerID * TCBSize
          If Tasks[TaskPtr + _Delay] > 0 Then
            Tasks[TaskPtr + _Counter] = Tasks[TaskPtr + _Delay]
          Else
            Tasks[TaskPtr + _Counter] = Tasks[TaskPtr + _Period]
          EndIf
          If Tasks[TaskPtr + _Misc] = OSCT_DONT_START Then
            Tasks[TaskPtr] = OSTCB_CYCTMR_WAITING
          Else
            Tasks[TaskPtr] = OSTCB_CYCTMR_STOPPED
          EndIf
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSRESETCYCLICTIMER_ERROR, INVALID_TIMERID)
          $endif
        EndIf
      EndProc


'---------------------------------------------------------------------------------------------------
' Returns true if the cyclic timer is running

      Proc OSCycTmrRunning(TaskID As Word), Word
        Dim TaskPtr As OSPW0
        Result = False
        If TimerID > LastTask Then Return
        TaskPtr = TimerID * TCBSize
        If Tasks[TaskPtr] > OSTCB_CYCTMR_STOPPED Then Result = True
      EndProc

'---------------------------------------------------------------------------------------------------
'  Return command for cyclic timer routine

      Proc OS_CycReturn()
        If Tasks[CurrTaskPtr + _Misc] = OSCT_CONTINUOUS Then
          Tasks[CurrTaskPtr + _Timer] = Tasks[CurrTaskPtr + _Period] ' reload Cyclic Timer
          Tasks[CurrTaskPtr] = OSTCB_CYCTMR_WAITING
        Else
          Tasks[CurrTaskPtr] = OSTCB_CYCTMR_STOPPED
        EndIf
        SelectAndRunTask()
      EndProc

    $endif


'----------------------
' Semaphore Routines
'----------------------

'---------------------------------------------------------------------------------------------------
' Suspends the current task until the binary or counting semaphore referenced in "Event_"
' has been signalled (binary semaphore) or the semaphore count has a value > 0 (counting semaphore)
' or the TimeOut (ticks) has been elapsed.

' If the event is already signalled (binary) or the semaphore count has a value > 0 (counting) when
' the wait routine is called then the current task will continue (no task switching).

' After a "OS_WaitSemaphore" the semaphore will be cleared (binary) or the semaphore count value
' will be decremented by one (counting semaphore, provided the value is > 0).

' If the wait times out the task will be resumed with the "OS_TimeOut" flag set. If the
' event is signalled within the TimeOut ticks the task will be resumed with the
' "OS_TimeOut" flag cleared. Use "OSNO_TIMEOUT" as "timeOut" parameter when an infinite timeout
' is needed.

    $ifdef OSENABLE_EVENTS

    ' Need to modify create code to use GetEventID

      Proc OS_WaitSem(EventID As Word, pTimeOut As Word)
        Dim EventPtr As OSPW0
        If EventID <= LastEvent Then
          Tasks[CurrTaskPtr + _Misc] = Tasks[CurrTaskPtr + _Misc] & mi_CLEAR_TIMEOUT  ' Clear timeout flag
          CheckEvents()                                 ' Check events, it may already have signalled
          EventPtr = EventID * ECBSize                  ' Point to the event in the event table
          If ECBs[EventPtr + _SEMCOUNT] > 0 Then        ' If already signalled, dec count and get out
            Dec ECBs[EventPtr + _SEMCOUNT]              ' if it reached zero will be checked next
            Return                                      ' CheckEvents
          EndIf
          GetReturnAddress()
          Tasks[CurrTaskPtr + _EventID] = EventID       ' Add the eventID into the current tasks TCB
          Tasks[CurrTaskPtr + _State] = OSTCB_TASK_WAITING ' Change current task to Waiting
          Tasks[CurrTaskPtr + _Timer] = pTimeout        ' add the Timeout to the TCB
          SelectAndRunTask()                            ' Move on to the next task
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OS_WAITSEM_ERROR, INVALID_EVENTID)
        $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Adds a binary semaphore to the eventslist.
' The function returns the semaphore's number or "OSNO_EVENT" if the eventslist is full
' "InitialValue" is the initial value assigned to the semaphore (true = "signalled").

      Proc OSCreateBinSem(InitialValue As Word), Word
        Dim EventPtr As OSPW1
        Result = GetEventID()                           ' Get a new EventID
        If Result <> OSNO_EVENT Then                    ' If Good EventID returned
          EventPtr = Result * ECBSize                   ' point to the ECB in the events list
          ECBs[EventPtr] = OSECB_EVENT_BINARY           ' set up semaphore type
          If InitialValue IsTrue Then                   ' Can only have 2 states Signalled (True or False)
            ECBs[EventPtr + _SEMCOUNT] = True
          Else
            ECBs[EventPtr + _SEMCOUNT] = False
          EndIf
          ECBs[EventPtr + _SEMISR] = 0                  ' Clear ISR Flag
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSCREATEBINSEM_ERROR, LIST_FULL)
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Adds a counting semaphore to the eventslist.
' The function returns the semaphore's number or "OSNO_EVENT" if the eventslist is full
' "InitialValue" is the initial count value assigned to the semaphore.

      Proc OSCreateSem(InitialValue As Word], Word
        Dim EventPtr As OSPW1
        Result = GetEventID()                           ' Get new EventID
        If Result <> OSNO_EVENT  Then                   ' If Good eventID returned
          EventPtr = Result * ECBSize                   ' point to the ecb
          ECBs[EventPtr] = OSECB_EVENT_COUNTING         ' set up semaphore type
          ECBs[EventPtr + _SEMCOUNT] = InitialValue     ' set initial value
          ECBs[EventPtr + _SEMISR] = 0                  ' Clear the ISR flag
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSCREATESEM_ERROR, LIST_FULL)
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Sets the binary semaphore to "signalled"  or increments the semaphore count (counting semaphore)

      Proc OSSignalSemaphore(EventID As Word)
        Dim EventPtr As OSPW0
        If EventID <= LastEvent Then
          EventPtr = EventID * ECBSize
          If ECBs[EventPtr] = OSECB_EVENT_COUNTING Then
            Inc ECBs[EventPtr +_SEMCOUNT]              ' Counting semaphore so increment count
          Else
            ECBs[EventPtr +_SEMCOUNT] = 1              ' Binary or Message so set to 1
          EndIf
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSSIGNALSEMAPHORE_ERROR, INVALID_EVENTID)
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Sets semaphore to "not signalled" (binary semaphore) or count to zero (counting semaphore).
' Also pending signaling from interrupts are cleared.

      Proc OSClearSemaphore(EventID As Word)
        Dim EventPtr As OSPW0
        If Event <= LastEvent Then
          EventPtr = EventID * ECBSize
          ECBs[EventPtr + _SEMCOUNT] = 0
          $ifdef OSENABLE_INTERRUPTS
    ' clear pending signaling from ISR's and disable Interrupts could change under interrupt
            Disable_Interrupts
            ECBs[EventPtr + _SEMISR] = 0
            Enable_Interrupts
          $endif
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSCLEARSEMAPHORE_ERROR, INVALID_EVENTID)
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' To be used from within interrupt service routines.
' Sets the semaphore to "signalled" (binary semaphore) or increments the semaphore count by one
' (counting semaphore)

      Proc OSSignalSemaphore_ISR(EventID As Word)      ' OSQueue_Event
        Dim EventPtrISR As Word                        ' Note must not be a shared variable
        If EventID <= LastEvent Then                   ' Check its a valid event number
          EventPtrISR = EventID * ECBSize              ' Point to relevent Event ECB
          Inc ECBs[EventPtrISR + _SEMISR]              ' Queue the ISR event
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' If Counting semaphore returns with the count value of the semaphore
' If Binary semaphore returns true if Signalled
' Action of reading Semaphore clears the signalled state

      Proc OSReadSemaphore(EventID As Word), Word
        Dim EventPtr As OSPW0
        Result = 0
        If EventID <= LastEvent Then
          CheckEvents()
          EventPtr = EventID * ECBSize
          Result = ECBs[EventPtr + _SEMCOUNT]
          ECBs[EventPtr + _SEMCOUNT]= 0
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSREADSEMAPHORE_ERROR, INVALID_EVENTID)
          $endif         
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Returns true if event is signalled, the event value remains unchanged

      Proc OSTrySemaphore(EventID As Word), Word
        Dim EventPtr As OSPW0
        Result = False
        If EventID <= LastEvent Then
          CheckEvents()
          EventPtr = EventID * ECBSize
          If ECBs[EventPtr + _SEMCOUNT] > 0 Then Result = True
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSTRYSEMAPHORE_ERROR, INVALID_EVENTID)
          $endif         
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Deletes an Event and frees up an ECB slot in the Events list providing no task is waiting on the
' event. Sets the eventID back to OSNO_EVENT.

      Proc OSDestroyEvent(ByRef EventId As Word)
        Dim EventIdx As OSPW0
        Dim NextTask As OSPW1

        EventIdx = Ptr16(EventID)
        If EventIdx <= LastEvent Then                      ' if its a valid EventID
          NextTask = 0
          Do                                               ' Check if any task is waiting on event
          If Tasks[NextTask] > OSTCB_DESTROYED Then        ' Skip empty or destroyed TCBs
            If Tasks[NextTask + _EventID] = EventIdx Then  ' Is it connected to an event
              If Tasks[NextTask] = OSTCB_TASK_WAITING Then ' and is it waiting on it
                $ifdef OSENABLE_ERROR_HANDLER
                  OS_RaiseError(OSDESTROYEVENT_ERROR, TASK_WAITING_ON_EVENT)
                $else
                  Return
                $endif
              EndIf
            EndIf
          EndIf
          NextTask = NextTask + TCBSize                     ' on to next task
          Loop Until NextTask >= LastTask
          ' to reach here there must be no tasks waiting on the event.
          ECBs[EventIdx] = OSECB_EVENT_NOEVENT              ' set the ecb to empty
          ECBs[EventIdx + _SEMCOUNT] = 0                    ' Clear to not signalled
          ECBs[EventIdx + _SEMISR] = 0                      ' and ISR queue
          Ptr16(EventID) = OSNO_EVENT                       ' clear the EventID to no event
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSDESTROYEVENT_ERROR, INVALID_EVENTID)
        $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Returns "true" if a timeout occured, or false if no timeout occurred (e.g. after "OS_WaitSemaphore").
' After calling "OS_TimeOut" the timeout flag is cleared.
' The flag is also initially cleared when an "OS_Wait..." routine is called.

      Proc OSTimeout(), Word
        Result = False
        If Tasks[CurrTaskPtr + _Misc] & $0100 = $0100 Then Result = True
        Tasks[CurrTaskPtr + _Misc] = Tasks[CurrTaskPtr + _Misc] & mi_CLEAR_TIMEOUT
      EndProc

'---------------------------
' Inter Task Messages
'---------------------------

' Intertask messaging works on the back of the binary semaphore mechanism
' Messages are passed as pointers to a variable. Both sender and receiver task must be aware of
' what type and size of data is to be passed.

'---------------------------------------------------------------------------------------------------
' OS_WaitMessage - If message is signalled will return with Address of message in WREG2,3
'                  If not signalled will yield to scheduler
'                  If Timeout was <> OSNO_TIMEOUT check Timeout flag before processing message

      Proc OS_WaitMessage(MessageID As Word, TimeOut As Word)
        Dim EventPtr As OSPW0
        If MessageID <= LastEvent Then
          Tasks[CurrTaskPtr + _Misc] = Tasks[CurrTaskPtr + _Misc] & mi_CLEAR_TIMEOUT ' Clear timeout flag
          CheckEvents()                                 ' Check events, it may already be signalled
          EventPtr = MessageID * ECBSize                ' Point to the event in the event table
          If ECBs[EventPtr + _SEMCOUNT] > 0 Then        ' If signalled,
            Dec ECBs[EventPtr + _SEMCOUNT]              ' Dec count, if it reached zero
            Return                                      ' will be checked next CheckEvents
          EndIf
          GetReturnAddress()
          Tasks[CurrTaskPtr + _EventID] = MessageID     ' Add the eventID into the current tasks TCB
          Tasks[CurrTaskPtr + _State] = OSTCB_TASK_WAITING ' Change current task to Waiting
          Tasks[CurrTaskPtr + _Timer] = Timeout         ' add the Timeout to the TCB
          SelectAndRunTask()                            ' Move on to the next task
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OS_WAITMESSAGE_ERROR, INVALID_MESSAGEID) '
        $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Create an event ECB entry for a message - MessageAddr should the variable location of the message
' or OSNO_MESSAGE if there is no message ready to send at time of creation
' Returns with the MessageID

      Proc OSCreateMessage(ByRef MessageAddr As Dword), Word
        Dim EventPtr As OSPW1
        Result = GetEventID()                          ' Get new MessageID
        If Result <> OSNO_EVENT Then                   ' if valid MessageID
          EventPtr = LastEvent * ECBSize               ' Point to actual offset into ECB Array
          ECBs[EventPtr] = OSECB_EVENT_MESSAGE         ' Set Event type
          If MessageAddr <> OSNO_MESSAGE Then          ' If message already ready
            ECBs[EventPtr + _SEMCOUNT] = 1             ' Set SemCount to Signalled
            ECBs[EventPtr + _MSGPTR_H] = MessageAddr.HighWord ' Set Upper Address bits of start of message
            ECBs[EventPtr + _MSGPTR_L] = MessageAddr.LowWord  ' Set Lower Address bits of start of message
          Else
            ECBs[EventPtr + _SEMCOUNT] = 0             ' Set SemCount to 0 - Unsignalled
          EndIf
          $ifdef OSENABLE_ERROR_HANDLER
          Else
            OS_RaiseError(OSCREATEMESSAGE_ERROR, LIST_FULL) ' Cant get a Message ID so raise error
          $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Returns a pointer to the current message associated with MessageID.  Returns 0 is there is no
' message.

      Proc OSReadMessage(MessageID As Word), WREG12.Dword
        Dim EventPtr As OSPW0
        Result = 0
        If MessageID <= LastEvent Then
          EventPtr = MessageID * ECBSize
          Result.HighWord = ECBs[EventPtr + _MSGPTR_H]' Load return value (WREG2,3) with
          Result.LowWord = ECBs[EventPtr + _MSGPTR_L] ' the address of message
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSREADMESSAGE_ERROR, INVALID_MESSAGEID) ' Invalid Message ID so raise error
        $endif
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Signals a message is ready to be read by a receiving task. MessageAddr is the address of the
' variable holding the message.

      Proc OSSignalMessage(MessageID As Word, ByRef MessageAddr As Dword)
        Dim EventPtr As OSPW0
        If MessageID <= LastEvent Then
          EventPtr = MessageID * ECBSize
          ECBs[EventPtr + _SEMCOUNT] = 1
          ECBs[EventPtr + _MSGPTR_H] = MessageAddr.HighWord
          ECBs[EventPtr + _MSGPTR_L] = MessageAddr.LowWord
        $ifdef OSENABLE_ERROR_HANDLER
        Else
          OS_RaiseError(OSSIGNALMESSAGE_ERROR, INVALID_MESSAGEID) ' Invalid Message ID so raise error
        $endif
        EndIf
      EndProc

    $endif

'--------------------------------------
' USART Interrupt Routines
'--------------------------------------

    $ifdef OSENABLE_SERIO_INTERRUPT

'---------------------------------------------------------------------------------------------------
' Waits until TX data buffer becomes empty.
      $if OSENABLE_TX_INTERRUPT = True | OSENABLE_BUFFERED_SERIO = True
        $define OS_WaitTXEmpty() OS_WaitSem(e_TXEvent, 0)
      $endif

'---------------------------------------------------------------------------------------------------
' Waits on RXChar being received. Preceeding data is held in buffer and can be accessed via HrsIn
' Not available unless buffered Serial IO is enabled

      $if OSENABLE_BUFFERED_SERIO = True
        Proc OS_WaitRXDataOn(RXChar As Byte, Timeout As Word)
          RXWaitFor = RSChar & $00FF
          OS_WaitSem(e_RXEvent, Timeout)
          RXWaitFor = OSNO_WAITFOR
        EndProc
      $endif

'---------------------------------------------------------------------------------------------------
' Waits on data input from USART. If data is buffered HrsIn will access the buffered data.

      $if OSENABLE_RX_INTERRUPT = True | OSENABLE_BUFFERED_SERIO = True
        $define OS_WaitRXData(Timeout) OS_WaitSem(e_RXEvent, Timeout)
      $endif

    $endif

'----------------------
' Exception Handling
'----------------------

'---------------------------------------------------------------------------------------------------
' Yields back to the OS with an Exception.  FuncCode defines the source of the exception, ErrCode
' defines the nature of the error. On Yielding the OS will run the error handler ahead of all other
' tasks.

' Error Handler is a Task created during initialisation and set to state of OSTCB_Task_Waiting.  
' It must have the highest priority to ensure that it gets run directly after the error is raised

' If a task doesn't yield to the OS when the OS detects timeout, The TimerISR will....
'   Record the ID of the running task and hold it in the Lower byte of the LastError variable.
'   Make the Error Handler task Eligible to Run
'   Adjust the Stack pointer back and push the address of the scheduler onto the stack.
' When ISR returns it returns to the scheduler rather than the blocking task.

' The Scheduler runs the highest priority eligible task - i.e. the error handler.

' Running the scheduler will reset the Timeout

'---------------------------------------------------------------------------------------------------
    $ifdef OSENABLE_ERROR_HANDLER

      Proc OSErrorHdlr_Init()
        ' Create the Error task, top priority - will jump to error handler if no TCB space
        t_Error = OSCreateTask(OSErrorTask, 0, False)
        Clear LastError                              ' reset lastError
      EndProc

'---------------------------------------------------------------------------------------------------
' This procedure is used both internally to raise errors also available for use within user tasks
' to handle error exceptions

      Proc OS_RaiseError(Source As Byte, ErrorCode As Byte)
        Dim TaskPtr As OSPW0
        LastError.HighByte = Source                  ' Set up the souce and 
        LastError.LowByte = ErrorCode                ' error codes
        If OSState = True Then                       ' If OS has been started
          GetReturnAddress()                         ' We can treat this as a yield to the OS
          TaskPtr = t_Error * TCBSize                ' calc position for task in task table
          Tasks[TaskPtr] = OSTCB_TASK_ELIGIBLE       ' set the erroring task to run at top priority 
          SelectAndRunTask()                         ' Select and run error routine as a task
        Else                                         ' If OS not started then
          GoSub OSErrorTask                          ' vall the Error handler as a subroutine
        EndIf
      EndProc

'---------------------------------------------------------------------------------------------------
' Error Handler Task - This task runs once on the occurence of an error and will either
' run User's Error Procedure or the default.
' Clears Last Error and

      OSErrorTask:
      Do
      $ifdef OSUSER_ERROR_HANDLER
        UserErrorProc()                              ' User procedure to handle errors
      $else
        DefaultErrorProc()                           ' OS Default handler if no User handler defined
      $endif
        Clear LastError                              ' Clear LastError
        If OSState = True Then                       ' OS has been started
          OS_Stop()                                  ' Stop the error handler and return to scheduler
        Else                                         ' if OS not started
          Return                                     ' back to application
        EndIf
      Loop
      
      Proc DefaultErrorProc()
        If OSState = True Then OS_WaitTXEmpty()
        HRSOut "\rOriginator $", Hex2 LastError.HighByte, " Error $", Hex2 LastError.LowByte, "\r"     
      EndProc
        

    $endif

 OverPRTOS:                   
